/* listv.h     v0.1  (C) 1999 adolfo@di-mare.com */

#ifndef   LIST_H
#define   LIST_H

/*-----------------------------------------------------*\
|  This is another list implementation, using a vector  |
|  instead of pointers and nodes. It doesn't come with  |
|  a matching "list.c" file, because all the list       |
|  operations are inlined, using the preprocessor.      |
|                                                       |
|  - I wrote this to show you that the iterators can    |
|    be used with different list implementations.       |
|                                                       |
|  - A vector list always has a fixed capacity of       |
|    "list_dim" elements. No boundary checks are        |
|    included for this implementation.                  |
|  - The do { ... } while (0) trick is useful because   |
|    you can create a block scope where variables get   |
|    defined, and also it compiles ok when a semicolon  |
|    is put after the inlined statement.                |
|  - Recall that list_insert(&L, p, &val) will insert   |
|    AFTER position "p". Hence, this operation cannot   |
|    be implemented as a simple                         |
|       memcpy(p, val, _L->_node_size);                 |
|  - I use a local variable "_L" to avoid evaluating    |
|    macro argument "L" more than once, but if the      |
|    macro must evaluate into an expression, this       |
|    isn't possible.                                    |
|  - I know this turns to be more complicated than it   |
|    should; I just couldn't resist the temptation to   |
|    pick on your brains!                               |
|                                                       |
|  To compile, just rename this file to "list.h", and   |
|  don't forget to avoid compiling and linking "list.c" |
|                                                       |
|  - If you want to learn more, try implementing this   |
|    same list, but make type "lpos" into an integer,   |
|    that is, a postion "p" would be an index into      |
|    array L->_first[p].                                |
|        How do you represent the NULL postion?         |
\*-----------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

#include <mem.h>

typedef struct { char nada; } *lpos;

#define list_dim 50 /* maximum capacity for any list */

typedef struct {
    lpos   _first;       /* private */
    size_t _count;       /* private */
    size_t _node_size;   /* private */
} list, *plist;

/* Note: on init() list size is fixed to "list_dim" elements */

#define list_init(L, VTYPE)                                   \
    do { list *_L = &(L);                                     \
        _L->_count     = 0;                                   \
        _L->_node_size = sizeof(VTYPE);                       \
        _L->_first = (lpos) malloc(list_dim * sizeof(VTYPE)); \
        memset(_L->_first, 0xf,list_dim * sizeof(VTYPE));     \
    } while (0)

#define list_done(L) free((L)->_first)

#define list_count(L) ((L)->_count)

/* Inserts "*val" AFTER position "p" in list "*L" */
#define list_insert(L, p, val)                                \
    do { list *L_ = (L);                                      \
         char *to  = (char*)(p);                              \
                                                              \
        if (NULL == to) {                                     \
            memmove(                                          \
                (char*)L_->_first + L_->_node_size,           \
                (char*)L_->_first,                            \
                L_->_node_size * L_->_count);                 \
            memcpy(L_->_first, val, L_->_node_size);          \
        } else {                                              \
            size_t chunk;                                     \
            to    = to + L_->_node_size;                      \
            chunk = (size_t)(to - (char*)L_->_first);         \
            chunk = (L_->_count * L_->_node_size) - chunk ;   \
            memmove(                                          \
                to + L_->_node_size,                          \
                to,                                           \
                chunk);                                       \
            memcpy(to, val, L_->_node_size);                  \
        }                                                     \
        L_->_count++;                                         \
    } while (0)

#define list_append(L, val)                      \
    do { list *_L = (L);                         \
        list_insert(_L, list_last(_L), val);     \
    } while (0)

#define list_first(L) \
    ((lpos)(0 == (L)->_count ? NULL : (L)->_first))

#define list_last(L)                                             \
   ((lpos) (                                                     \
   (   0==(L)->_count                                            \
   ?  NULL                                                       \
   :  (char*)((L)->_first) + (L)->_node_size * ((L)->_count-1)   \
   ) ))

#define list_next(L,p)                            \
        ( p == list_last(L) ? NULL :              \
          (lpos) ((char*)(p) + (L)->_node_size) )

/* convert to (void) pointer to stored value */
#define list_retrieve(L,p) ((void*) (p))

#ifdef __cplusplus
}
#endif

#endif /* LIST_H */
/* EOF:  listv.h */
